import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
DB_PATH = os.path.join(ROOT_DIR, '../../dat/db')
TREE_PATH = os.path.join(ROOT_DIR, '../../dat/tree')
INFO_PATH = os.path.join(ROOT_DIR, '../../dat/info')
INFO_FNAME = "info.json"
